# Copyright© 2025-2026 Gesellschaft zur Förderung der angewandten Forschung e.V.
# acting on behalf of its Fraunhofer Institut für Graphische Datenverarbeitung.
# Licensed under the EUPL. See LICENSE.txt.

import cupy as cp
from .trainsum import TrainSum

trainsum = TrainSum[cp.ndarray](cp)
